@@ -81,6 +81,7 @@ enum {
 	PROP_PROTOCOL,
 	PROP_STATE,
 	PROP_MAX_INCOMING_PAYLOAD_SIZE,
+	PROP_KEEPALIVE_INTERVAL,
 };
 
 enum {
@@ -107,6 +108,7 @@ struct _SoupWebsocketConnectionPrivate {
 	char *origin;
 	char *protocol;
 	guint64 max_incoming_payload_size;
+	guint keepalive_interval;
 
 	gushort peer_close_code;
 	char *peer_close_data;
@@ -131,11 +133,13 @@ struct _SoupWebsocketConnectionPrivate {
 	/* Current message being assembled */
 	guint8 message_opcode;
 	GByteArray *message_data;
+
+	GSource *keepalive_timeout;
 };
 
 #define MAX_INCOMING_PAYLOAD_SIZE_DEFAULT   128 * 1024
 
-G_DEFINE_TYPE (SoupWebsocketConnection, soup_websocket_connection, G_TYPE_OBJECT)
+G_DEFINE_TYPE_WITH_PRIVATE (SoupWebsocketConnection, soup_websocket_connection, G_TYPE_OBJECT)
 
 typedef enum {
 	SOUP_WEBSOCKET_QUEUE_NORMAL = 0,
@@ -162,8 +166,7 @@ soup_websocket_connection_init (SoupWebsocketConnection *self)
 {
 	SoupWebsocketConnectionPrivate *pv;
 
-	pv = self->pv = G_TYPE_INSTANCE_GET_PRIVATE (self, SOUP_TYPE_WEBSOCKET_CONNECTION,
-						     SoupWebsocketConnectionPrivate);
+	pv = self->pv = soup_websocket_connection_get_instance_private (self);
 
 	pv->incoming = g_byte_array_sized_new (1024);
 	g_queue_init (&pv->outgoing);
@@ -224,6 +227,18 @@ stop_output (SoupWebsocketConnection *self)
 	}
 }
 
+static void
+keepalive_stop_timeout (SoupWebsocketConnection *self)
+{
+	SoupWebsocketConnectionPrivate *pv = self->pv;
+
+	if (pv->keepalive_timeout) {
+		g_source_destroy (pv->keepalive_timeout);
+		g_source_unref (pv->keepalive_timeout);
+		pv->keepalive_timeout = NULL;
+	}
+}
+
 static void
 close_io_stop_timeout (SoupWebsocketConnection *self)
 {
@@ -241,6 +256,7 @@ close_io_stream (SoupWebsocketConnection *self)
 {
 	SoupWebsocketConnectionPrivate *pv = self->pv;
 
+	keepalive_stop_timeout (self);
 	close_io_stop_timeout (self);
 
 	if (!pv->io_closing) {
@@ -331,6 +347,11 @@ send_message (SoupWebsocketConnection *self,
 	guint8 *mask = 0;
 	guint8 *at;
 
+	if (!(soup_websocket_connection_get_state (self) == SOUP_WEBSOCKET_STATE_OPEN)) {
+		g_debug ("Ignoring message since the connection is closed or is closing");
+		return;
+	}
+
 	bytes = g_byte_array_sized_new (14 + length);
 	outer = bytes->data;
 	outer[0] = 0x80 | opcode;
@@ -411,6 +432,8 @@ send_close (SoupWebsocketConnection *self,
 
 	send_message (self, flags, 0x08, (guint8 *)buffer, len);
 	self->pv->close_sent = TRUE;
+
+	keepalive_stop_timeout (self);
 }
 
 static void
@@ -511,6 +534,59 @@ too_big_error_and_close (SoupWebsocketConnection *self,
 	stop_input (self);
 }
 
+static void
+close_connection (SoupWebsocketConnection *self,
+                  gushort                  code,
+                  const char              *data)
+{
+	SoupWebsocketQueueFlags flags;
+	SoupWebsocketConnectionPrivate *pv;
+
+	pv = self->pv;
+
+	if (pv->close_sent) {
+		g_debug ("close code already sent");
+		return;
+	}
+
+	/* Validate the closing code received by the peer */
+	switch (code) {
+	case SOUP_WEBSOCKET_CLOSE_NORMAL:
+	case SOUP_WEBSOCKET_CLOSE_GOING_AWAY:
+	case SOUP_WEBSOCKET_CLOSE_PROTOCOL_ERROR:
+	case SOUP_WEBSOCKET_CLOSE_UNSUPPORTED_DATA:
+	case SOUP_WEBSOCKET_CLOSE_BAD_DATA:
+	case SOUP_WEBSOCKET_CLOSE_POLICY_VIOLATION:
+	case SOUP_WEBSOCKET_CLOSE_TOO_BIG:
+		break;
+	case SOUP_WEBSOCKET_CLOSE_NO_EXTENSION:
+		if (pv->connection_type == SOUP_WEBSOCKET_CONNECTION_SERVER) {
+			g_debug ("Wrong closing code %d received for a server connection",
+			         code);
+			break;
+		}
+	case SOUP_WEBSOCKET_CLOSE_SERVER_ERROR:
+		if (pv->connection_type != SOUP_WEBSOCKET_CONNECTION_SERVER) {
+			g_debug ("Wrong closing code %d received for a non server connection",
+			         code);
+			break;
+		}
+	default:
+		g_debug ("Wrong closing code %d received", code);
+	}
+
+	g_signal_emit (self, signals[CLOSING], 0);
+
+	if (pv->close_received)
+		g_debug ("responding to close request");
+
+	flags = 0;
+	if (pv->connection_type == SOUP_WEBSOCKET_CONNECTION_SERVER && pv->close_received)
+		flags |= SOUP_WEBSOCKET_QUEUE_LAST;
+	send_close (self, flags, code, data);
+	close_io_after_timeout (self);
+}
+
 static void
 receive_close (SoupWebsocketConnection *self,
 	       const guint8 *data,
@@ -542,8 +618,7 @@ receive_close (SoupWebsocketConnection *self,
 		if (pv->connection_type == SOUP_WEBSOCKET_CONNECTION_SERVER)
 			close_io_stream (self);
 	} else {
-		/* Send back the response */
-		soup_websocket_connection_close (self, pv->peer_close_code, NULL);
+		close_connection (self, pv->peer_close_code, NULL);
 	}
 }
 
@@ -586,6 +661,7 @@ process_contents (SoupWebsocketConnection *self,
 			receive_ping (self, payload, payload_len);
 			break;
 		case 0x0A:
+			g_debug ("received pong message");
 			break;
 		default:
 			g_debug ("received unsupported control frame: %d", (int)opcode);
@@ -831,6 +907,12 @@ on_web_socket_output (GObject *pollable_stream,
 	gssize count;
 	gsize len;
 
+	if (soup_websocket_connection_get_state (self) == SOUP_WEBSOCKET_STATE_CLOSED) {
+		g_debug ("Ignoring message since the connection is closed");
+		stop_output (self);
+		return TRUE;
+	}
+
 	frame = g_queue_peek_head (&pv->outgoing);
 
 	/* No more frames to send */
@@ -997,6 +1079,10 @@ soup_websocket_connection_get_property (GObject *object,
 		g_value_set_uint64 (value, pv->max_incoming_payload_size);
 		break;
 
+	case PROP_KEEPALIVE_INTERVAL:
+		g_value_set_uint (value, pv->keepalive_interval);
+		break;
+
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 		break;
@@ -1041,6 +1127,11 @@ soup_websocket_connection_set_property (GObject *object,
 		pv->max_incoming_payload_size = g_value_get_uint64 (value);
 		break;
 
+	case PROP_KEEPALIVE_INTERVAL:
+		soup_websocket_connection_set_keepalive_interval (self,
+		                                                  g_value_get_uint (value));
+		break;
+
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 		break;
@@ -1079,6 +1170,7 @@ soup_websocket_connection_finalize (GObject *object)
 	g_assert (pv->io_closing);
 	g_assert (pv->io_closed);
 	g_assert (!pv->close_timeout);
+	g_assert (!pv->keepalive_timeout);
 
 	if (pv->message_data)
 		g_byte_array_free (pv->message_data, TRUE);
@@ -1096,8 +1188,6 @@ soup_websocket_connection_class_init (SoupWebsocketConnectionClass *klass)
 {
 	GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
 
-	g_type_class_add_private (klass, sizeof (SoupWebsocketConnectionPrivate));
-
 	gobject_class->constructed = soup_websocket_connection_constructed;
 	gobject_class->get_property = soup_websocket_connection_get_property;
 	gobject_class->set_property = soup_websocket_connection_set_property;
@@ -1227,6 +1317,26 @@ soup_websocket_connection_class_init (SoupWebsocketConnectionClass *klass)
 							      G_PARAM_CONSTRUCT |
 							      G_PARAM_STATIC_STRINGS));
 
+	/**
+	 * SoupWebsocketConnection:keepalive-interval:
+	 *
+	 * Interval in seconds on when to send a ping message which will
+	 * serve as a keepalive message. If set to 0 the keepalive message is
+	 * disabled.
+	 *
+	 * Since: 2.58
+	 */
+	g_object_class_install_property (gobject_class, PROP_KEEPALIVE_INTERVAL,
+					 g_param_spec_uint ("keepalive-interval",
+					                    "Keepalive interval",
+					                    "Keepalive interval",
+					                    0,
+					                    G_MAXUINT,
+					                    0,
+					                    G_PARAM_READWRITE |
+					                    G_PARAM_CONSTRUCT |
+					                    G_PARAM_STATIC_STRINGS));
+
 	/**
 	 * SoupWebsocketConnection::message:
 	 * @self: the WebSocket
@@ -1574,7 +1684,6 @@ soup_websocket_connection_close (SoupWebsocketConnection *self,
 				 gushort code,
 				 const char *data)
 {
-	SoupWebsocketQueueFlags flags;
 	SoupWebsocketConnectionPrivate *pv;
 
 	g_return_if_fail (SOUP_IS_WEBSOCKET_CONNECTION (self));
@@ -1589,16 +1698,7 @@ soup_websocket_connection_close (SoupWebsocketConnection *self,
 	else
 		g_return_if_fail (code != SOUP_WEBSOCKET_CLOSE_SERVER_ERROR);
 
-	g_signal_emit (self, signals[CLOSING], 0);
-
-	if (pv->close_received)
-		g_debug ("responding to close request");
-
-	flags = 0;
-	if (pv->connection_type == SOUP_WEBSOCKET_CONNECTION_SERVER && pv->close_received)
-		flags |= SOUP_WEBSOCKET_QUEUE_LAST;
-	send_close (self, flags, code, data);
-	close_io_after_timeout (self);
+	close_connection (self, code, data);
 }
 
 /**
@@ -1646,3 +1746,68 @@ soup_websocket_connection_set_max_incoming_payload_size (SoupWebsocketConnection
 		g_object_notify (G_OBJECT (self), "max-incoming-payload-size");
 	}
 }
+
+/**
+ * soup_websocket_connection_get_keepalive_interval:
+ * @self: the WebSocket
+ *
+ * Gets the keepalive interval in seconds or 0 if disabled.
+ *
+ * Returns: the keepalive interval.
+ *
+ * Since: 2.58
+ */
+guint
+soup_websocket_connection_get_keepalive_interval (SoupWebsocketConnection *self)
+{
+	SoupWebsocketConnectionPrivate *pv;
+
+	g_return_val_if_fail (SOUP_IS_WEBSOCKET_CONNECTION (self), 0);
+	pv = self->pv;
+
+	return pv->keepalive_interval;
+}
+
+static gboolean
+on_queue_ping (gpointer user_data)
+{
+	SoupWebsocketConnection *self = SOUP_WEBSOCKET_CONNECTION (user_data);
+
+	g_debug ("sending ping message");
+	send_message (self, SOUP_WEBSOCKET_QUEUE_NORMAL, 0x09, NULL, 0);
+
+	return G_SOURCE_CONTINUE;
+}
+
+/**
+ * soup_websocket_connection_set_keepalive_interval:
+ * @self: the WebSocket
+ * @interval: the interval to send a ping message or 0 to disable it
+ *
+ * Sets the interval in seconds on when to send a ping message which will serve
+ * as a keepalive message. If set to 0 the keepalive message is disabled.
+ *
+ * Since: 2.58
+ */
+void
+soup_websocket_connection_set_keepalive_interval (SoupWebsocketConnection *self,
+                                                  guint                    interval)
+{
+	SoupWebsocketConnectionPrivate *pv;
+
+	g_return_if_fail (SOUP_IS_WEBSOCKET_CONNECTION (self));
+	pv = self->pv;
+
+	if (pv->keepalive_interval != interval) {
+		pv->keepalive_interval = interval;
+		g_object_notify (G_OBJECT (self), "keepalive-interval");
+
+		keepalive_stop_timeout (self);
+
+		if (interval > 0) {
+			pv->keepalive_timeout = g_timeout_source_new_seconds (interval);
+			g_source_set_callback (pv->keepalive_timeout, on_queue_ping, self, NULL);
+			g_source_attach (pv->keepalive_timeout, pv->main_context);
+		}
+	}
+}
